#include "deluge.h"


char *
urlescape(char *s, int n)
{
	static char buf[512];
	int i;
	char *p;
	char c;

	if(n < 0)
		n = strlen(s);

	p = buf;
	for(i = 0; i < n; i++){
		c = s[i];
		if(c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z' || c >= '0' && c <= '9')
			*p++ = c;
		else
			p = seprint(p, buf+sizeof buf, "%%%02X", (int)(uchar)s[i]);
	}
	*p = 0;
	return buf;
}


char *
urlparse(char *url, int *https, char **host, char **port, char **path)
{
	char *p, *ep;

	p = url;

	*https = 0;
	if(strncmp(p, "http://", strlen("http://")) == 0){
		p += strlen("http://");
	}else if(strncmp(p, "https://", strlen("https://")) == 0){
		*https = 1;
		p += strlen("https://");
	}else
		return smprint("missing http:// or https://");

	ep = p;
	while(*ep != '/' && *ep != ':')
		ep++;
	*host = emallocz((ep - p) + 1, 1);
	memmove(*host, p, ep-p);
	if(*ep == ':'){
		p = ep+1;
		ep = p;
		while(*ep >= '0' && *ep <= '9')
			ep++;
		*port = emallocz((ep - p) + 1, 1);
		memmove(*port, p, ep-p);
	}else
		*port = estrdup(*https ? "443" : "80");

	*path = estrdup(ep);

	return nil;
}

 
char *
trackerget(Bee *b, Trackreq *tr)
{
	int fd, sfd;
	String *qs;
	Biobuf *bp, bpbuf;
	char *p;
	char buf[64];
	char *l;
	int code;
	char *errmsg;
	int c;
	int https;
	char *host, *port, *path;
	TLSconn conn;

	host = port = path = nil;
	errmsg = urlparse(tr->announce, &https, &host, &port, &path);
	if(errmsg) {
		free(host); free(port); free(path);
		return errmsg;
	}

	bp = &bpbuf;
	DEBUG(2,"tracker info: http://%s:%s/%s\n",host,port,path);
	fd = dial(netmkaddr(host, nil, port), 0, 0, 0);
	if(fd < 0) {
		free(host); free(port); free(path);
		return smprint("dial %s: %r", netmkaddr(host, nil, port));
	}

	if(https){
		memset(&conn, 0, sizeof conn);
		sfd = tlsClient(fd, &conn);
		if(sfd < 0){
			close(fd);
			free(host); free(port); free(path);
			return smprint("tlsclient: %r");
		}
		if(conn.cert)
				free(conn.cert);
		close(fd);
		fd = sfd;
	}

	qs = s_new();
	s_append(qs, "info_hash=");
	s_append(qs, urlescape((char*)tr->infohash, 20));
	s_append(qs, "&peer_id=");
	s_append(qs, urlescape((char*)tr->peerid, 20));
	s_append(qs,"&key=");
	s_append(qs,urlescape((char*)tr->keyid,Keysize));
	s_append(qs, "&port=");
	s_append(qs, urlescape(tr->listenport, -1));
	s_append(qs, "&uploaded=");
	snprint(buf, sizeof buf, "%lld", tr->ul);
	s_append(qs, urlescape(buf, -1));
	s_append(qs, "&downloaded=");
	snprint(buf, sizeof buf, "%lld", tr->dl);
	s_append(qs, urlescape(buf, -1));
	s_append(qs, "&left=");
	snprint(buf, sizeof buf, "%lld", tr->left);
	s_append(qs, urlescape(buf, -1));
	s_append(qs,"&compact=1");

	if(tr->event && tr->event[0]){
		s_append(qs, "&event=");
		s_append(qs, urlescape(tr->event, -1));
	}

	DEBUG(2,
		"GET %s?%s HTTP/1.0\r\n"
		"Host: %s\r\n"
		"User-agent: Plan9/deluge\r\n\r\n",
		path, s_to_c(qs), host);

	fprint(fd,
		"GET %s?%s HTTP/1.0\r\n"
		"Host: %s\r\n"
		"User-agent: Plan9/deluge\r\n\r\n",
		path, s_to_c(qs), host);
	

	s_free(qs);
	free(host); free(port); free(path);

	if(Binit(bp, fd, OREAD) < 0)
		return smprint("Binit for http fd: %r");

	l = Brdline(bp, '\n');
	if(l == nil)
		return smprint("error reading response: %r");

	p = l + strlen("HTTP/1.0 ");
	 
	code = atoi(p);
	if(code != 200){
		Bterm(bp);
		close(fd);
		DEBUG(2,"tracker said: %s",l);
		return smprint("http get failed, code=%d\n", code);
	}

	/* skip through http headers */
	for(;;){
		l = Brdline(bp, (int)'\n');
		if(l == nil){
			Bterm(bp);
			close(fd);
			return smprint("error/unexpected eof while reading tracker response: %r\n");
		}
		if(Blinelen(bp) == 2 && strncmp(l, "\r\n", 2) == 0)
			break;
	}

	errmsg = beeunpickle(bp, b);
	while((c = Bgetc(bp)) >= 0){
		if(c == '\r' || c == '\n')
			continue;
		DEBUG(2, "extra data: ");
		while((c = Bgetc(bp)) >= 0)
			DEBUG(2, "%c", c);
		DEBUG(2, "\n");
		Bterm(bp);
		close(fd);
		return smprint("data after bencode data from tracker (%d)", c);
	}
	Bterm(bp);
	close(fd);
	if(errmsg)
		return smprint("invalid bencode from tracker: %s", errmsg);
	return nil;
}
